// fdm83.c - finite-difference frequency-domain cochlear model
//           MOH83

#include <stdio.h>
#include <math.h>

typedef struct {float x, y;} complex;

complex a[8][8][512], p[8][512], y[512], d[512];
float   w, p0p, srd, dx, dy, r1, r2, r4;
float   xl, yh;
int     m, n;

static float rho = 1;
static float tpi = 6.283183;

complex
cmplx(float x, float y)
{
    complex z;

    z.x = x;
    z.y = y;
    return (z);
}

complex
c_neg(complex z)
{
    complex y;

    y.x = -z.x;
    y.y = -z.y;
    return (y);
}

complex
c_recip(complex z)
{
    complex y;
    float   msq;

    msq = z.x * z.x + z.y * z.y;
    y.x = z.x / msq;
    y.y = -z.y / msq;
    return (y);
}

double
cabs(complex z)
{
    double   msq;

    msq = z.x * z.x + z.y * z.y;
    return (sqrt(msq));
}

complex
c_add(complex a, complex b)
{
    complex c;

    c.x = a.x + b.x;
    c.y = a.y + b.y;
    return (c);
}

complex
c_sub(complex a, complex b)
{
    complex c;

    c.x = a.x - b.x;
    c.y = a.y - b.y;
    return (c);
}

complex
c_mul(complex a, complex b)
{
    complex c;

    c.x = a.x * b.x - a.y * b.y;
    c.y = a.x * b.y + a.y * b.x;
    return (c);
}

complex
c_div(complex a, complex b)
{
    complex c;
    float msq;

    msq = b.x * b.x + b.y * b.y;
    c.x = (a.x * b.x + a.y * b.y) / msq;
    c.y = (a.y * b.x - a.x * b.y) / msq;
    return (c);
}

// mxint - initialize pressure and zero first sub-matrix

void
mxint(int m, int n)
{
    complex p1;
    int     i, j;

    p1 = cmplx(-2. * dx * p0p, 0.0);
    for (i = 0; i < m; i++) {
	p[i][0] = p1;
	p[i][n] = cmplx(0.0, 0.0);
	for (j = 0; j < m; j++)
	    a[i][j][0] = cmplx(0.0, 0.0);
    }
}

// mxadd - initialize the kth sub-matrix and add the previously inverted 
//	  (k-1)th sub-matrix:
//
//         a(k) = a(k) + a(k-1)

void
mxadd(int k, int m)
{
    int     i, m1, m2;

    m1 = m - 1;
    m2 = m - 2;
    a[0][0][k] = c_add(a[0][0][k], cmplx(r4 - y[k].y * srd, y[k].x * srd));
    a[0][1][k].x -= r2;
    for (i = 1; i < m1; i++) {
	a[i][i - 1][k].x -= r1;
	a[i][i][k].x += r4;
	a[i][i + 1][k].x -= r1;
    } 
    a[m1][m2][k].x -= r2;
    a[m1][m1][k].x += r4;
}

// mxinv - invert the kth sub-matrix of a(i,j) and solve for the pressure.

void
mxinv(int k, int m)
{
    complex at, b[8][8], q[8];
    int i, ii, j, ji, j1, mm;

    // work down from top of a(k)
    mm = m - 1;
    for (j = 0; j < m; j++) {
        if(cabs(a[j][j][k]) < 1e-9) {
	    fprintf(stderr, "*** Ill-conditioned matrix!\n");
            exit(1);
	}
        at = c_recip(a[j][j][k]);		    // at = 1/a(i,j,k)
        for (j1 = 0; j1 < m; j1++) {
            a[j][j1][k] = c_mul(a[j][j1][k], at);    // a(i,j1,k)=a(j,j1,k)*at
        }
        a[j][j][k] = at;
        if (j < mm) {
	    for (i = j + 1; i < m; i++) {
                at = a[i][j][k];
                a[i][j][k] = cmplx(0.0, 0.0);
	        for (j1 = 0; j1 < m; j1++) {
                    a[i][j1][k] = c_sub(a[i][j1][k], c_mul(a[j][j1][k], at));
                }
            }
        }
    }
   // work up from bottom
    for (i = 0; i < m; i++) {
        for (j = 0; j < m; j++) {
            b[i][j] = a[i][j][k];
	}
     }
    for (i = 0; i < mm; i++) {
        for (j = i + 1; j < m; j++) {
            b[i][j] = cmplx(0.0, 0.0);
        }
    }
    for (ii = 0; ii < mm; ii++) {
        i = m - ii - 2;
	for (ji = i + 1; ji < m; ji++) {
	    for (j = 0; j < m; j++) {
                b[i][j] = c_sub(b[i][j], c_mul(a[i][ji][k], b[ji][j]));
	    }
	}
    }

    // solve p(k) = b*p(k) and a(k) = -b
    for (i = 0; i < m; i++) {
        at = cmplx(0.0, 0.0);
	for (j = 0; j < m; j++) {
          at = c_add(at, c_mul(p[j][k], b[i][j]));
          a[i][j][k] = c_neg(b[i][j]);
        }
        q[i] = at;
    }
    for (i = 0; i < m; i++)
        p[i][k] = q[i];

}


// mxdbl - double the value of a(k)

void
mxdbl(int k, int m)
{
    int i, j;

    for (i = 0; i < m; i++)
        for (j = 0; j < m; j++)
            a[i][j][k] = c_add(a[i][j][k], a[i][j][k]);

}

// mxcpy - copy the (k-1)th submatrix to the kth submatrix

void
mxcpy(int k, int m)
{
    int i, j, km;

    km = k - 1;
    for (i = 0; i < m; i++) {
        p[i][k] = p[i][km];
        for (j = 0; j < m; j++) {
	    a[i][j][k] = a[i][j][km];
	}
    }

}

// mxelm - eliminate the upper-diagonal sub-matrices:
//
//         p(k) = p(k) - a(k)*p(k+1)

void
mxelm(int k, int m)
{
    complex ap;
    int i, j, kp;

    kp = k + 1;
    for (i = 0; i < m; i++) {
        ap = cmplx(0.0, 0.0);
	for (j = 0; j < m; j++) {
            ap = c_add(ap, c_mul(a[i][j][k], p[j][kp]));
	}
        p[i][k] = c_sub(p[i][k], ap);
    }

}

// dbph - compute magnitude (in dB) and phase (unwrapped)

void
dbph(complex *p, int n)
{
    int i;
    double mg, db, ph, lp = -0.5;
    static double tpi = 2 * 3.1415927;
    
    for (i = 0; i < n; i++) {
        mg = hypot(p[i].x, p[i].y);
        if (mg > 1e-30) {
            db = 20 * log10(mg);
            ph = atan2(p[i].y, p[i].x) / tpi;
        } else {
            db = -600;
            ph = 0;
        }
        if (ph > lp) {
            ph -= floor(ph - lp + 0.5);
        } else {
            ph += floor(lp - ph + 0.5);
        }
        lp = ph;
        p[i].x = db;
        p[i].y = ph;
    }
}

//  wrsoln - write out data to disk

void
wrsoln(int ii, double f)
{
    char fn[80];
    complex tc;
    double dl, zr, zi;
    int i;
    FILE *fp;

    fprintf(stdout, "m=%d n=%d xl=%.2f yh=%.2f f=%.2f\n", m, n, xl, yh, f);
    sprintf(fn, "fdm83_%d.dat", ii);
    fp = fopen(fn, "w");
    fprintf(fp, "; %s - f=%.0f\n", fn, f);
    fprintf(fp, "; m=%d n=%d xl=%.2f yh=%.2f\n", m, n, xl, yh);

    /* BM displacement */
    tc = cmplx(0.0, -1 / w);
    for (i = 0; i < n; i++)
    	d[i] = c_mul(c_mul(y[i], p[0][i]),tc);

    /* convert to magnitude and phase */
    dbph(p[0], n);
    dbph(d, n);
    dbph(y, n);

    /* write data */
    fprintf(fp, "; x/L          P                 Dbm               Zbm     \n");
    fprintf(fp, ";         [dB]     [ph]     [dB]     [ph]     [re]     [im] \n");
    dl = 1.0 / (n - 1);
    for (i = 0; i < n; i++) {
	zr = pow(10.0, -0.05 * y[i].x) * cos(-2 * 3.1415927 * y[i].y);
	zi = pow(10.0, -0.05 * y[i].x) * sin(-2 * 3.1415927 * y[i].y);
        fprintf(fp, "%6.3f %8.1f %8.2f %8.1f %8.2f %8.0f %8.0f\n", i * dl,
        p[0][i].x, p[0][i].y, d[i].x, d[i].y, zr, zi);
    }
    fclose(fp);
}

/* c_krmx - compute impedance z(f,x) = k(x)/(iw) + r(x) + m(x)*(iw) 
 *          where w=2*pi*f
 *                k(x)=k0*exp(ke*x)
 *                r(x)=r0*exp(re*x)
 *                m(x)=m0*exp(me*x)
 */
complex
c_krmx(float ko, float ke, float ro, float re, float rc, float rd,
    float mo, float me, float x, float w)
{
    complex z;

    z.x = ro * exp (re * x) + rc * exp (rd * x);
    z.y = mo * exp (me * x) * w - ko * exp (ke * x) / w;

    return (z);
}

/* imped - compute partition impedance */

complex
imped(float x, float w)
{
    complex z1, z2, z3, za, zb;

    static float k1o = 1e9, k1e = -2.4;    /* stiffness component of Z1 */
    static float r1o = 400, r1e = 0;       /* damping component of Z1 */
    static float r1c = 1e3, r1d = -1.2;    /* r1=r1o*e^r1e*x + r1c*e^r1d*x */
    static float m1o = 0.001, m1e = 1.2;   /* mass component of Z1 */
    static float k2o = 250, k2e = -2.2;    /* stiffness component of Z2 */
    static float r2o = 1e-4, r2e = 1.1;    /* damping component of Z2 */
    static float r2c = 8e-4, r2d = -2.2;   /* r2=r2o*e^r2e*x + r2c*e^r2d*x */
    static float m2o = 3e-9, m2e = 2.2;    /* mass component of Z2 */
    static float k3o = 0, k3e = 0;         /* stiffness component of Z3 */
    static float r3o = 1, r3e = 0;         /* damping component of Z3 */

    /* Cochlear partition */

    z1 = c_krmx(k1o, k1e, r1o, r1e, r1c, r1d, m1o, m1e, x, w);
    z2 = c_krmx(k2o, k2e, r2o, r2e, r2c, r2d, m2o, m2e, x, w);
    z3 = c_krmx(k3o, k3e, r3o, r3e, 0, 0, 0, 0, x, w);
    /* za = z3 * z3 / z2 */
    za = c_mul(z3, c_div(z3, z2));
    /* zb = z1 - za */
    zb = c_sub(z1, za);

    return (zb);
}

complex
admit(float x, float w)
{
    complex z;
    static float rhe = 10;
    static float xhe = 2.5;

    if (x > xhe)
	z = cmplx(rhe, 0.0);
    else
	z = imped(x, w);

    return (c_recip(z));
}

int
main()
{
    float   ho, f, x;
    int     nm, k;

    xl = 2.55;
    yh = 0.1;
    m = 6;
    n = 409;

    dx = xl/(n-1);
    dy = yh/(m-1);
    r1 = (dx/dy)*(dx/dy);
    r2 = r1 + r1;
    r4 = 2. + r2;
    nm = n - 1;

    for (ho = 0; ho < 7; ho++) {	// loop over ten frequencies

	f = 500*pow(2, ho);
	w = tpi*f;
	srd = 4*w*rho*dy*r1;
	p0p = 2*w*w*rho;

	for (k = 0; k < n; k++) {	// compute admittance for all x
	    x = dx * k;
	    y[k] = admit(x, w);
        }

	// solve block matrix equation

	mxint(m, n);
	mxadd(0, m);
	mxinv(0, m);
	mxdbl(0, m);
	for (k = 1; k < nm; k++) {		// work down from top
	    mxcpy(k, m);
	    mxadd(k, m);
	    mxinv(k, m);
	}

        // work up from bottom
        for (k = nm - 1;  k > 0; k--) {
	    mxelm(k, m);
        }
        wrsoln(ho, f);
    }

    return (0);
}
